#! /usr/bin/env python
# -*- coding: utf8 -*-

"""
aubio
=====

Provides a number of classes and functions for music and audio signal
analysis.

How to use the documentation
----------------------------

Documentation of the python module is available as docstrings provided
within the code, and a reference guide available online from `the
aubio homepage <https://aubio.org/documentation>`_.

The docstrings examples are written assuming `aubio` and `numpy` have been
imported with:

>>> import aubio
>>> import numpy as np
"""

import numpy
from ._aubio import __version__ as version
from ._aubio import float_type
from ._aubio import *
from .midiconv import *
from .slicing import *


class fvec(numpy.ndarray):
    """fvec(input_arg=1024)
    A vector holding float samples.

    If `input_arg` is an `int`, a 1-dimensional vector of length `input_arg`
    will be created and filled with zeros. Otherwise, if `input_arg` is an
    `array_like` object, it will be converted to a 1-dimensional vector of
    type :data:`float_type`.

    Parameters
    ----------
    input_arg : `int` or `array_like`
        Can be a positive integer, or any object that can be converted to
        a numpy array with :func:`numpy.array`.

    Examples
    --------
    >>> aubio.fvec(10)
    array([0., 0., 0., 0., 0., 0., 0., 0., 0., 0.], dtype=float32)
    >>> aubio.fvec([0,1,2])
    array([0., 1., 2.], dtype=float32)
    >>> a = np.arange(10); type(a), type(aubio.fvec(a))
    (<class 'numpy.ndarray'>, <class 'numpy.ndarray'>)
    >>> a.dtype, aubio.fvec(a).dtype
    (dtype('int64'), dtype('float32'))

    Notes
    -----

    In the Python world, `fvec` is simply a subclass of
    :class:`numpy.ndarray`. In practice, any 1-dimensional `numpy.ndarray` of
    `dtype` :data:`float_type` may be passed to methods accepting
    `fvec` as parameter. For instance, `sink()` or `pvoc()`.

    See Also
    --------
    cvec : a container holding spectral data
    numpy.ndarray : parent class of :class:`fvec`
    numpy.zeros : create a numpy array filled with zeros
    numpy.array : create a numpy array from an existing object
    """
    def __new__(cls, input_arg=1024):
        if isinstance(input_arg, int):
            if input_arg == 0:
                raise ValueError("vector length of 1 or more expected")
            return numpy.zeros(input_arg, dtype=float_type, order='C')
        else:
            np_input = numpy.array(input_arg, dtype=float_type, order='C')
            if len(np_input.shape) != 1:
                raise ValueError("input_arg should have shape (n,)")
            if np_input.shape[0] == 0:
                raise ValueError("vector length of 1 or more expected")
            return np_input



"""utility routines to slice sound files at given timestamps"""

import os
from aubio import source, sink

_max_timestamp = 1e120


def slice_source_at_stamps(source_file, timestamps, timestamps_end=None,
                           output_dir=None, samplerate=0, hopsize=256,
                           create_first=False):
    """Slice a sound file at given timestamps.

    This function reads `source_file` and creates slices, new smaller
    files each starting at `t` in `timestamps`, a list of integer
    corresponding to time locations in `source_file`, in samples.

    If `timestamps_end` is unspecified, the slices will end at
    `timestamps_end[n] = timestamps[n+1]-1`, or the end of file.
    Otherwise, `timestamps_end` should be a list with the same length
    as `timestamps` containing the locations of the end of each slice.

    If `output_dir` is unspecified, the new slices will be written in
    the current directory. If `output_dir` is a string, new slices
    will be written in `output_dir`, after creating the directory if
    required.

    The default `samplerate` is 0, meaning the original sampling rate
    of `source_file` will be used. When using a sampling rate
    different to the one of the original files, `timestamps` and
    `timestamps_end` should be expressed in the re-sampled signal.

    The `hopsize` parameter simply tells :class:`source` to use this
    hopsize and does not change the output slices.

    If `create_first` is True and `timestamps` does not start with `0`, the
    first slice from `0` to `timestamps[0] - 1` will be automatically added.

    Parameters
    ----------
    source_file : str
        path of the resource to slice
    timestamps : :obj:`list` of :obj:`int`
        time stamps at which to slice, in samples
    timestamps_end : :obj:`list` of :obj:`int` (optional)
        time stamps at which to end the slices
    output_dir : str (optional)
        output directory to write the slices to
    samplerate : int (optional)
        samplerate to read the file at
    hopsize : int (optional)
        number of samples read from source per iteration
    create_first : bool (optional)
        always create the slice at the start of the file

    Examples
    --------
    Create two slices: the first slice starts at the beginning of the
    input file `loop.wav` and lasts exactly one second, starting at
    sample `0` and ending at sample `44099`; the second slice starts
    at sample `44100` and lasts until the end of the input file:

    >>> aubio.slice_source_at_stamps('loop.wav', [0, 44100])

    Create one slice, from 1 second to 2 seconds:

    >>> aubio.slice_source_at_stamps('loop.wav', [44100], [44100 * 2 - 1])

    Notes
    -----
    Slices may be overlapping. If `timestamps_end` is `1` element
    shorter than `timestamps`, the last slice will end at the end of
    the file.
    """

    if not timestamps:
        raise ValueError("no timestamps given")

    if timestamps[0] != 0 and create_first:
        timestamps = [0] + timestamps
        if timestamps_end is not None:
            timestamps_end = [timestamps[1] - 1] + timestamps_end

    if timestamps_end is not None:
        if len(timestamps_end) == len(timestamps) - 1:
            timestamps_end = timestamps_end + [_max_timestamp]
        elif len(timestamps_end) != len(timestamps):
            raise ValueError("len(timestamps_end) != len(timestamps)")
    else:
        timestamps_end = [t - 1 for t in timestamps[1:]] + [_max_timestamp]

    regions = list(zip(timestamps, timestamps_end))

    source_base_name, _ = os.path.splitext(os.path.basename(source_file))
    if output_dir is not None:
        if not os.path.isdir(output_dir):
            os.makedirs(output_dir)
        source_base_name = os.path.join(output_dir, source_base_name)

    def _new_sink_name(source_base_name, timestamp, samplerate):
        # create name based on a timestamp in samples, converted in seconds
        timestamp_seconds = timestamp / float(samplerate)
        return source_base_name + "_%011.6f" % timestamp_seconds + '.wav'

    # open source file
    _source = source(source_file, samplerate, hopsize)
    samplerate = _source.samplerate

    total_frames = 0
    slices = []

    while True:
        # get hopsize new samples from source
        vec, read = _source.do_multi()
        # if the total number of frames read will exceed the next region start
        while regions and total_frames + read >= regions[0][0]:
            # get next region
            start_stamp, end_stamp = regions.pop(0)
            # create a name for the sink
            new_sink_path = _new_sink_name(source_base_name, start_stamp,
                                           samplerate)
            # create its sink
            _sink = sink(new_sink_path, samplerate, _source.channels)
            # create a dictionary containing all this
            new_slice = {'start_stamp': start_stamp, 'end_stamp': end_stamp,
                         'sink': _sink}
            # append the dictionary to the current list of slices
            slices.append(new_slice)

        for current_slice in slices:
            start_stamp = current_slice['start_stamp']
            end_stamp = current_slice['end_stamp']
            _sink = current_slice['sink']
            # sample index to start writing from new source vector
            start = max(start_stamp - total_frames, 0)
            # number of samples yet to written be until end of region
            remaining = end_stamp - total_frames + 1
            # not enough frames remaining, time to split
            if remaining < read:
                if remaining > start:
                    # write remaining samples from current region
                    _sink.do_multi(vec[:, start:remaining], remaining - start)
                    # close this file
                    _sink.close()
            elif read > start:
                # write all the samples
                _sink.do_multi(vec[:, start:read], read - start)
        total_frames += read
        # remove old slices
        slices = list(filter(lambda s: s['end_stamp'] > total_frames,
                             slices))
        if read < hopsize:
            break



# -*- coding: utf-8 -*-
""" utilities to convert midi note number to and from note names """

import sys
from ._aubio import freqtomidi, miditofreq

__all__ = ['note2midi', 'midi2note', 'freq2note', 'note2freq']

py3 = sys.version_info[0] == 3
if py3:
    str_instances = str
    int_instances = int
else:
    str_instances = (str, unicode)
    int_instances = (int, long)


def note2midi(note):
    """Convert note name to midi note number.

    Input string `note` should be composed of one note root
    and one octave, with optionally one modifier in between.

    List of valid components:

    - note roots: `C`, `D`, `E`, `F`, `G`, `A`, `B`,
    - modifiers: `b`, `#`, as well as unicode characters
      `𝄫`, `♭`, `♮`, `♯` and `𝄪`,
    - octave numbers: `-1` -> `11`.

    Parameters
    ----------
    note : str
        note name

    Returns
    -------
    int
        corresponding midi note number

    Examples
    --------
    >>> aubio.note2midi('C#4')
    61
    >>> aubio.note2midi('B♭5')
    82

    Raises
    ------
    TypeError
        If `note` was not a string.
    ValueError
        If an error was found while converting `note`.

    See Also
    --------
    midi2note, freqtomidi, miditofreq
    """
    _valid_notenames = {'C': 0, 'D': 2, 'E': 4, 'F': 5, 'G': 7,
                        'A': 9, 'B': 11}
    _valid_modifiers = {
            u'𝄫': -2,                         # double flat
            u'♭': -1, 'b': -1, '\u266d': -1,  # simple flat
            u'♮': 0, '\u266e': 0, None: 0,    # natural
            '#': +1, u'♯': +1, '\u266f': +1,  # sharp
            u'𝄪': +2,                         # double sharp
            }
    _valid_octaves = range(-1, 10)
    if not isinstance(note, str_instances):
        msg = "a string is required, got {:s} ({:s})"
        raise TypeError(msg.format(str(type(note)), repr(note)))
    if len(note) not in range(2, 5):
        msg = "string of 2 to 4 characters expected, got {:d} ({:s})"
        raise ValueError(msg.format(len(note), note))
    notename, modifier, octave = [None] * 3

    if len(note) == 4:
        notename, modifier, octave_sign, octave = note
        octave = octave_sign + octave
    elif len(note) == 3:
        notename, modifier, octave = note
        if modifier == '-':
            octave = modifier + octave
            modifier = None
    else:
        notename, octave = note

    notename = notename.upper()
    octave = int(octave)

    if notename not in _valid_notenames:
        raise ValueError("%s is not a valid note name" % notename)
    if modifier not in _valid_modifiers:
        raise ValueError("%s is not a valid modifier" % modifier)
    if octave not in _valid_octaves:
        raise ValueError("%s is not a valid octave" % octave)

    midi = (octave + 1) * 12 + _valid_notenames[notename] \
                             + _valid_modifiers[modifier]
    if midi > 127:
        raise ValueError("%s is outside of the range C-2 to G8" % note)
    return midi


def midi2note(midi):
    """Convert midi note number to note name.

    Parameters
    ----------
    midi : int [0, 128]
        input midi note number

    Returns
    -------
    str
        note name

    Examples
    --------
    >>> aubio.midi2note(70)
    'A#4'
    >>> aubio.midi2note(59)
    'B3'

    Raises
    ------
    TypeError
        If `midi` was not an integer.
    ValueError
        If `midi` is out of the range `[0, 128]`.

    See Also
    --------
    note2midi, miditofreq, freqtomidi
    """
    if not isinstance(midi, int_instances):
        raise TypeError("an integer is required, got %s" % midi)
    if midi not in range(0, 128):
        msg = "an integer between 0 and 127 is excepted, got {:d}"
        raise ValueError(msg.format(midi))
    _valid_notenames = ['C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#',
                        'A', 'A#', 'B']
    return _valid_notenames[midi % 12] + str(int(midi / 12) - 1)


def freq2note(freq):
    """Convert frequency in Hz to nearest note name.

    Parameters
    ----------
    freq : float [0, 23000[
        input frequency, in Hz

    Returns
    -------
    str
        name of the nearest note

    Example
    -------
    >>> aubio.freq2note(440)
    'A4'
    >>> aubio.freq2note(220.1)
    'A3'
    """
    nearest_note = int(freqtomidi(freq) + .5)
    return midi2note(nearest_note)


def note2freq(note):
    """Convert note name to corresponding frequency, in Hz.

    Parameters
    ----------
    note : str
        input note name

    Returns
    -------
    freq : float [0, 23000[
        frequency, in Hz

    Example
    -------
    >>> aubio.note2freq('A4')
    440
    >>> aubio.note2freq('A3')
    220.1
    """
    midi = note2midi(note)
    return miditofreq(midi)
